# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test email validation."""

import unittest

from mailman.interfaces.address import (
    IEmailValidator,
    InvalidEmailAddressError,
)
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility


class TestValidate(unittest.TestCase):
    """Test email validation."""

    layer = ConfigLayer

    def setUp(self):
        self.is_valid = getUtility(IEmailValidator).is_valid
        self.validate = getUtility(IEmailValidator).validate

    def test_no_at(self):
        self.assertFalse(self.is_valid('anne'))

    def test_no_domain(self):
        self.assertFalse(self.is_valid('anne@'))

    def test_valid_email(self):
        self.assertTrue(self.is_valid('anne@example.com'))

    def test_two_ats(self):
        self.assertTrue(self.is_valid('anne@x@example.com'))

    def test_valid_email_quoted_local(self):
        self.assertTrue(self.is_valid('"anne"@example.com'))

    def test_two_ats_quoted_local(self):
        self.assertTrue(self.is_valid('"anne@x"@example.com'))

    def test_validator_exception(self):
        self.assertRaises(InvalidEmailAddressError,
                          self.validate,
                          'bad_address')



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Tests for the Task runner."""

import os
import unittest

from datetime import timedelta
from lazr.config import as_timedelta
from mailman.app.lifecycle import create_list
from mailman.app.moderator import hold_message
from mailman.config import config
from mailman.database.transaction import dbconnection
from mailman.interfaces.cache import ICacheManager
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.pending import IPendable, IPendings
from mailman.interfaces.requests import IListRequests
from mailman.interfaces.workflow import IWorkflowStateManager
from mailman.model.bounce import BounceEvent
from mailman.model.message import Message
from mailman.runners.task import TaskRunner
from mailman.testing.helpers import (
    LogFileMark,
    make_testable_runner,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from mailman.utilities.datetime import factory
from zope.component import getUtility
from zope.interface import implementer


@implementer(IPendable)
class MyPendable(dict):
    PEND_TYPE = 'my pended'


class TestTask(unittest.TestCase):
    """Test various aspects of the Task runner."""

    layer = ConfigLayer

    def setUp(self):
        self._pendings = getUtility(IPendings)
        self._events = BounceEvent
        self._wfmanager = getUtility(IWorkflowStateManager)
        self._cachemanager = getUtility(ICacheManager)
        self._messages = getUtility(IMessageStore)
        self._mlist = create_list('ant@example.com')
        self._msg1 = mfs("""\
To: ant@example.com
From: anne@example.com
Subject: A message
Message-ID: <msg1>

first message
""")
        self._msg2 = mfs("""\
To: ant@example.com
From: anne@example.com
Subject: A message
Message-ID: <msg2>

second message
""")
        self._listrequests = IListRequests(self._mlist)
        self._runner = make_testable_runner(TaskRunner)
        self._cachemanager.add('cache1', 'xxx1', lifetime=timedelta(days=1))
        self._cachemanager.add('cache2', 'xxx2', lifetime=timedelta(days=3))
        pendable = MyPendable(id=1)
        self._token1 = self._pendings.add(pendable, lifetime=timedelta(days=1))
        self._token2 = self._pendings.add(pendable, lifetime=timedelta(days=3))
        self._wfmanager.save(self._token1)
        self._wfmanager.save(self._token2)
        self._requestid1 = hold_message(
            self._mlist, self._msg1, reason='Testing')
        self._requestid2 = hold_message(
            self._mlist, self._msg2, reason='Testing')

    def test_task_runner(self):
        # Test that the task runner deletes expired cache, pendings and
        # associated workflows.
        self.assertEqual(self._cachemanager.get('cache1'), 'xxx1')
        self.assertEqual(self._cachemanager.get('cache2'), 'xxx2')
        self.assertEqual(self._pendings.count(), 4)
        self.assertEqual(self._wfmanager.count, 2)
        mark = LogFileMark('mailman.task')
        factory.fast_forward(days=2)
        self._runner.run()
        self.assertIsNone(self._cachemanager.get('cache1'))
        self.assertEqual(self._cachemanager.get('cache2'), 'xxx2')
        self.assertEqual(self._pendings.count(), 3)
        pended = self._pendings.confirm(self._token2, expunge=False)
        self.assertEqual(pended['type'], 'my pended')
        self.assertEqual(list(self._wfmanager.get_all_tokens()),
                         [self._token2])
        log = mark.read()
        self.assertIn('Task runner evicted 1 expired pendings', log)
        self.assertIn('Task runner deleted 1 orphaned workflows', log)
        self.assertIn('Task runner deleted 0 orphaned requests', log)
        self.assertIn('Task runner evicted expired cache entries', log)

    def test_task_runner_request(self):
        # Test that the task runner deletes orphaned requests.
        self.assertEqual(self._pendings.count(), 4)
        self.assertEqual(self._listrequests.count, 2)
        life = as_timedelta(config.mailman.moderator_request_life)
        mark = LogFileMark('mailman.task')
        factory.fast_forward(days=life.days+1)
        self._runner.run()
        self.assertEqual(self._pendings.count(), 0)
        self.assertEqual(self._listrequests.count, 0)
        log = mark.read()
        self.assertIn('Task runner deleted 2 orphaned requests', log)

    def test_task_runner_messages(self):
        # Test that the task runner deletes orphaned messages from the
        # message store.
        # Initially, there are 2 messages in the store and 4 pendings.
        self.assertEqual(len(list(self._messages.messages)), 2)
        self.assertEqual(self._pendings.count(), 4)
        # Deleting the first request removes the pending but not the message.
        self._listrequests.delete_request(self._requestid1)
        self.assertEqual(self._pendings.count(), 3)
        self.assertEqual(len(list(self._messages.messages)), 2)
        mark = LogFileMark('mailman.task')
        self._runner.run()
        # Now there's only msg2.
        self.assertEqual(len(list(self._messages.messages)), 1)
        self.assertIsNotNone(self._messages.get_message_by_id('<msg2>'))
        log = mark.read()
        self.assertIn('Task runner deleted 1 orphaned messages', log)

    @dbconnection
    def test_task_runner_message_files(self, store):
        # Test that task runner deletes message files with no entry in the
        # message store.
        def count_files():
            """ A helper to count the number of saved message files."""
            count = 0
            base_dir = config.MESSAGES_DIR
            for root, dirs, files in os.walk(base_dir):
                if files:
                    count += len(files)
            return count
        # Initally there are two entries in the message store and two files.
        self.assertEqual(len(list(self._messages.messages)), 2)
        self.assertEqual(count_files(), 2)
        # Delete one of the message store entries leaving one but still two
        # files.
        row = store.query(Message).filter_by(message_id='<msg1>').first()
        store.delete(row)
        self.assertEqual(len(list(self._messages.messages)), 1)
        self.assertEqual(count_files(), 2)
        mark = LogFileMark('mailman.task')
        self._runner.run()
        # Now there's only one file.
        self.assertEqual(len(list(self._messages.messages)), 1)
        self.assertEqual(count_files(), 1)
        log = mark.read()
        self.assertIn('Task runner deleted 1 orphaned message files', log)

    @dbconnection
    def test_task_runner_bounce_events_old_unprocessed(self, store):
        # Test that the task runner deletes processed bounce events older than
        # dsn_lifetime, but not newer ones or unprocessed ones.
        # Set one old but unprocessed.
        event = self._events(self._mlist.list_id,
                             'anne@example.com',
                             self._msg1,
                             )
        event.timestamp -= (as_timedelta(config.mailman.dsn_lifetime) +
                            as_timedelta('1d'))
        store.add(event)
        mark = LogFileMark('mailman.task')
        self._runner.run()
        log = mark.read()
        self.assertIn('Task runner evicted 0 expired bounce events', log)

    @dbconnection
    def test_task_runner_bounce_events_old_processed(self, store):
        # Test that the task runner deletes processed bounce events older than
        # dsn_lifetime, but not newer ones or unprocessed ones.
        # Set one old and processed.
        event = self._events(self._mlist.list_id,
                             'anne@example.com',
                             self._msg1,
                             )
        event.timestamp -= (as_timedelta(config.mailman.dsn_lifetime) +
                            as_timedelta('1d'))
        event.processed = True
        store.add(event)
        mark = LogFileMark('mailman.task')
        self._runner.run()
        log = mark.read()
        self.assertIn('Task runner evicted 1 expired bounce events', log)

    @dbconnection
    def test_task_runner_bounce_events_processed(self, store):
        # Test that the task runner deletes processed bounce events older than
        # dsn_lifetime, but not newer ones or unprocessed ones.
        # Set one processed.
        event = self._events(self._mlist.list_id,
                             'anne@example.com',
                             self._msg1,
                             )
        event.processed = True
        store.add(event)
        mark = LogFileMark('mailman.task')
        self._runner.run()
        log = mark.read()
        self.assertIn('Task runner evicted 0 expired bounce events', log)



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the string utilities."""

import unittest

from mailman.testing.layers import ConfigLayer
from mailman.utilities import string


class TestString(unittest.TestCase):
    def test_oneline_bogus_charset(self):
        self.assertEqual(string.oneline('foo', 'bogus'), 'foo')

    def test_wrap_blank_paragraph(self):
        self.assertEqual(string.wrap('\n\n'), '\n\n')


class TestWebSubstitutions(unittest.TestCase):
    """Test that all the web substitutions are available."""

    layer = ConfigLayer

    # Create a mock MailingList object

    class MockMailingList:

        def __init__(self, domain, list_id, mail_host):

            class preferred_language:

                def __init__(self, code):
                    self.code = code

            self.domain = domain
            self.list_id = list_id
            self.mail_host = mail_host
            self.fqdn_listname = list_id.replace('.', '@', 1)
            self.display_name = 'MyList'
            self.list_name = 'mylist'
            self.short_listname = 'mylist'
            self.description = 'description'
            self.info = 'info'
            self.request_address = list_id.replace('.', '-request@', 1)
            self.owner_address = list_id.replace('.', '-owner@', 1)
            self.preferred_language = preferred_language('en')

    # Create a mock domain object
    class MockDomain:

        def __init__(self, base_url):
            self.base_url = base_url

    def test_web_urls(self):
        base_url = "https://example.com/mailman3"
        list_id = "mylist.lists.example.com"
        mail_host = "lists.example.com"
        held_message_url = "https://example.com/mailman3/lists/mylist.lists.example.com/held_messages"  # noqa: E501
        mailing_list_url = "https://example.com/mailman3/lists/mylist.lists.example.com"  # noqa: E501
        domain_url = "https://example.com/mailman3/domains/lists.example.com"
        pending_subscriptions_url = "https://example.com/mailman3/lists/mylist.lists.example.com/subscription_requests"   # noqa: E501
        pending_unsubscriptions_url = "https://example.com/mailman3/lists/mylist.lists.example.com/unsubscription_requests"  # noqa: E501

        # Expected URLs are generated.
        domain = self.MockDomain(base_url)
        mlist = self.MockMailingList(domain, list_id, mail_host)
        # Create a template with all the substitutions.
        template = """\
$base_url
$list_id
$mail_host
$held_message_url
$mailing_list_url
$domain_url
$pending_subscriptions_url
$pending_unsubscriptions_url
"""
        template = string.expand(template, mlist)
        # Assert the expected values
        self.assertIn(base_url, template)
        self.assertIn(list_id, template)
        self.assertIn(mail_host, template)
        self.assertIn(held_message_url, template)
        self.assertIn(mailing_list_url, template)
        self.assertIn(domain_url, template)
        self.assertIn(pending_subscriptions_url, template)
        self.assertIn(pending_unsubscriptions_url, template)



# Copyright (C) 2001-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Task runner."""

import os
import logging

from datetime import datetime
from lazr.config import as_timedelta
from mailman.config import config
from mailman.core.runner import Runner
from mailman.database.transaction import dbconnection, transactional
from mailman.interfaces.cache import ICacheManager
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.pending import IPendings
from mailman.interfaces.workflow import IWorkflowStateManager
from mailman.model.bounce import BounceEvent
from mailman.model.requests import _Request
from mailman.utilities.datetime import now
from public import public
from zope.component import getUtility


dlog = logging.getLogger('mailman.debug')
tlog = logging.getLogger('mailman.task')


@public
class TaskRunner(Runner):
    """The task runner."""

    is_queue_runner = False

    def __init__(self, name, slice=None):
        super().__init__(name, slice)
        self.lastrun = datetime.min
        self.lastmsg = datetime.min
        self.delay = as_timedelta(config.mailman.run_tasks_every)
        # Only look for orphaned message files this often.  Normally there
        # shouldn't be any anyway.
        self.msgdelay = as_timedelta('1w')

    @transactional
    def _do_periodic(self):
        """Invoked periodically by the run() method in the super class."""
        if self.lastrun + self.delay > datetime.now():
            return                                    # pragma: nocover
        self.lastrun = datetime.now()
        dlog.debug('Running task runner periodic tasks')
        self._evict_pendings()
        self._evict_expired_bounce_events()
        self._evict_cache()

    @dbconnection
    def _get_requests(self, store):
        # Get (id, token) for all requests.
        results = store.query(_Request).all()
        yield from [(result.id, result.data_hash) for result in results]

    @dbconnection
    def _delete_request(self, store, id):
        # Delete the request with id = id.
        request = (store.query(_Request).get(id))
        if request is not None:
            store.delete(request)

    def _evict_pendings(self):
        pendings = getUtility(IPendings)
        wfmanager = getUtility(IWorkflowStateManager)
        opendings = pendings.count()
        owflows = wfmanager.count
        pendings.evict()
        count = opendings - pendings.count()
        tlog.info('Task runner evicted %d expired pendings', count)
        # Now delete any orphaned workflow states.
        for token in wfmanager.get_all_tokens():
            if pendings.confirm(token, expunge=False) is None:
                wfmanager.discard(token)
        count = owflows - wfmanager.count
        tlog.info('Task runner deleted %d orphaned workflows', count)
        count = 0
        for id, token in self._get_requests():
            if pendings.confirm(token, expunge=False) is None:
                self._delete_request(id)
                count += 1
        tlog.info('Task runner deleted %d orphaned requests', count)
        # Also, delete any orphaned messages and message files from the
        # message store. The order of the next 3 steps is important to
        # avoid premature deletion due to race conditions.
        # First get a list of message files, but only do this infrequently.
        message_files = []
        if self.lastmsg + self.msgdelay <= datetime.now():
            self.lastmsg = datetime.now()
            for root, dirs, files in os.walk(config.MESSAGES_DIR):
                if files:
                    for f in files:
                        message_files.append(os.path.join(root, f))
        # Then a list of message-ids in the store and a dict of their hashes.
        messages = getUtility(IMessageStore)
        msgs = []
        hashes = dict()
        for msg in messages.messages:
            if msg is not None:
                msgs.append(msg.get('message-id'))
                hashes[msg['message-id-hash']] = True
        # Finally, a list of pending message-ids.
        pmids = dict()
        for token, pendable in pendings:
            if not pendable:
                continue                                # pragma: nocover
            mid = pendable.get('_mod_message_id')
            if mid:
                pmids[mid] = True
        # Now delete the orphaned message store messages.
        count = 0
        for mmid in msgs:
            if mmid not in pmids:
                messages.delete_message(mmid)
                count += 1
        tlog.info('Task runner deleted %d orphaned messages', count)
        # And finally the orphaned message files
        count = 0
        for message_file in message_files:
            if os.path.basename(message_file) not in hashes:
                os.remove(message_file)
                count += 1
        tlog.info('Task runner deleted %d orphaned message files', count)

    @dbconnection
    def _evict_expired_bounce_events(self, store):
        count = 0
        for entry in store.query(BounceEvent).all():
            if not entry.processed:
                continue
            if entry.timestamp > now() - as_timedelta(
                    config.mailman.dsn_lifetime):
                continue
            store.delete(entry)
            count += 1
        tlog.info('Task runner evicted %d expired bounce events', count)

    def _evict_cache(self):
        getUtility(ICacheManager).evict_expired()
        tlog.info('Task runner evicted expired cache entries')



# Copyright (C) 2020-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""add_domain_webhost

Revision ID: 729a8f2045da
Revises: 2156fc3f6f7d
Create Date: 2024-03-23 20:40:52.500161

"""
import mailman
import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = '729a8f2045da'
down_revision = '2156fc3f6f7d'


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('domain', schema=None) as batch_op:
        batch_op.add_column(sa.Column(
            'base_url', mailman.database.types.SAUnicode(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('domain', schema=None) as batch_op:
        batch_op.drop_column('base_url')

    # ### end Alembic commands ###



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test master watcher utilities."""

import os
import time
import signal
import socket
import tempfile
import unittest

from click.testing import CliRunner
from contextlib import ExitStack, suppress
from datetime import timedelta
from flufl.lock import Lock, TimeOutError
from importlib.resources import path
from io import StringIO
from mailman.bin import master
from mailman.config import config
from mailman.testing.helpers import LogFileMark, TestableMaster
from mailman.testing.layers import ConfigLayer
from unittest.mock import patch


class FakeLock:
    details = ('host.example.com', 9999, '/tmp/whatever')

    def unlock(self):
        pass


class TestMaster(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        fd, self.lock_file = tempfile.mkstemp()
        os.close(fd)
        # The lock file should not exist before we try to acquire it.
        os.remove(self.lock_file)

    def tearDown(self):
        # Unlocking removes the lock file, but just to be safe (i.e. in case
        # of errors).
        with suppress(FileNotFoundError):
            os.remove(self.lock_file)

    def test_acquire_lock_1(self):
        lock = master.acquire_lock_1(False, self.lock_file)
        is_locked = lock.is_locked
        lock.unlock()
        self.assertTrue(is_locked)

    def test_acquire_lock_1_force(self):
        # Create the lock and lock it.
        my_lock = Lock(self.lock_file)
        my_lock.lock(timedelta(seconds=60))
        # Try to aquire it again with force.
        lock = master.acquire_lock_1(True, self.lock_file)
        self.assertTrue(lock.is_locked)
        lock.unlock()

    def test_master_state(self):
        my_lock = Lock(self.lock_file)
        # Mailman is not running.
        state, lock = master.master_state(self.lock_file)
        self.assertEqual(state, master.WatcherState.none)
        # Acquire the lock as if another process had already started the
        # master.  Use a timeout to avoid this test deadlocking.
        my_lock.lock(timedelta(seconds=60))
        try:
            state, lock = master.master_state(self.lock_file)
        finally:
            my_lock.unlock()
        self.assertEqual(state, master.WatcherState.conflict)

    def test_master_state_stale(self):
        # Create a lock file with non-existent pid.
        with open(self.lock_file, 'w') as fp:
            fp.write(f'{self.lock_file}|{socket.getfqdn()}|9999999|junk')
        # Try to acquire the lock.
        Lock(self.lock_file)
        state, lock = master.master_state(self.lock_file)
        self.assertEqual(state, master.WatcherState.stale_lock)

    def test_acquire_lock_timeout_reason_unknown(self):
        stderr = StringIO()
        with ExitStack() as resources:
            resources.enter_context(patch(
                'mailman.bin.master.acquire_lock_1',
                side_effect=TimeOutError))
            resources.enter_context(patch(
                'mailman.bin.master.master_state',
                return_value=(master.WatcherState.none, FakeLock())))
            resources.enter_context(patch(
                'mailman.bin.master.sys.stderr', stderr))
            with self.assertRaises(SystemExit) as cm:
                master.acquire_lock(False)
            self.assertEqual(cm.exception.code, 1)
            self.assertEqual(stderr.getvalue(), """\
For unknown reasons, the master lock could not be acquired.

Lock file: {}
Lock host: host.example.com

Exiting.
""".format(config.LOCK_FILE))

    def test_main_cli(self):
        command = CliRunner()
        fake_lock = FakeLock()
        with ExitStack() as resources:
            config_file = str(resources.enter_context(
                path('mailman.testing', 'testing.cfg')))
            init_mock = resources.enter_context(patch(
                'mailman.bin.master.initialize'))
            lock_mock = resources.enter_context(patch(
                'mailman.bin.master.acquire_lock',
                return_value=fake_lock))
            start_mock = resources.enter_context(patch.object(
                master.Loop, 'start_runners'))
            loop_mock = resources.enter_context(patch.object(
                master.Loop, 'loop'))
            command.invoke(
                master.main,
                ('-C', config_file,
                 '--no-restart', '--force',
                 '-r', 'in:1:1', '--verbose'))
            # We got initialized with the custom configuration file and the
            # verbose flag.
            init_mock.assert_called_once_with(config_file, True)
            # We returned a lock that was force-acquired.
            lock_mock.assert_called_once_with(True)
            # We created a non-restartable loop.
            start_mock.assert_called_once_with([('in', 1, 1)])
            loop_mock.assert_called_once_with()

    def test_sighup_handler(self):
        """Invokes the SIGHUP handler.

        This sends SIGHUPs to the runners.  Unfortunately, there is no
        easy way to test whether that actually happens.  We'd need a
        specialized runner for that.

        """
        m = TestableMaster()
        mark = LogFileMark('mailman.runner')
        m.start('command')
        self.assertEqual(len(list(m._kids)), 1)

        # We need to give the runner some time to install the signal
        # handler.  If we send SIGHUP before it does, the default
        # action is to terminate the process.  We wait until we see
        # that the runner has done so by inspecting the log file.
        # This is race free, and bounded in time.
        start = time.time()
        while ("runner started." not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)

        mark = LogFileMark('mailman.runner')
        m._sighup_handler(None, None)

        # Check if the runner reopened the log.
        start = time.time()
        needle = "command runner caught SIGHUP.  Reopening logs."
        while (needle not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)
        self.assertIn(needle, mark.read())

        # Just to make sure it didn't die.
        self.assertEqual(len(list(m._kids)), 1)
        m.stop()

    def test_sigusr1_handler(self):
        """Invokes the SIGUSR1 handler.

        We then check whether the runners restart.
        """
        m = TestableMaster()
        m._restartable = True
        mark = LogFileMark('mailman.runner')
        m.start('command')
        kids = list(m._kids)
        self.assertEqual(len(kids), 1)
        old_kid = kids[0]

        # We need to give the runner some time to install the signal
        # handler.  If we send SIGUSR1 before it does, the default
        # action is to terminate the process.  We wait until we see
        # that the runner has done so by inspecting the log file.
        # This is race free, and bounded in time.
        start = time.time()
        while ("runner started." not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)

        # Invoke the handler in a loop.  This is race free, and
        # bounded in time.
        start = time.time()
        mark = LogFileMark('mailman.runner')
        while old_kid in set(m._kids) and time.time() - start < 10:
            # We must not send signals in rapid succession, because
            # the behavior of signals arriving while the process is in
            # the signal handler varies.  Linux implements System V
            # semantics, which means the default signal action is
            # restored for the duration of the signal handler.  In
            # this case it means to terminate the process.
            time.sleep(1)
            m._sigusr1_handler(None, None)

        # Check if the runner got the signal.
        start = time.time()
        needle = "command runner caught SIGUSR1.  Stopping."
        while (needle not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)
        self.assertIn(needle, mark.read())

        new_kids = list(m._kids)
        self.assertEqual(len(new_kids), 1)
        self.assertTrue(kids[0] != new_kids[0])
        m._restartable = False
        for pid in new_kids:
            try:
                os.kill(pid, signal.SIGTERM)
            except ProcessLookupError:
                pass
        m.thread.join()
        m.cleanup()

    def test_sigterm_handler(self):
        """Invokes the SIGTERM handler.

        We then check whether the runners are actually stopped.
        """
        m = TestableMaster()
        mark = LogFileMark('mailman.runner')
        m.start('command')
        kids = list(m._kids)
        self.assertEqual(len(kids), 1)
        old_kid = kids[0]

        # We need to give the runner some time to install the signal
        # handler.  If we send SIGTERM before it does, the default
        # action is to terminate the process and will return a
        # slightly different status code.  We wait until we see that
        # the runner has done so by inspecting the log file.  This is
        # race free, and bounded in time.
        start = time.time()
        while ("runner started." not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)

        # Invoke the handler in a loop.  This is race free, and
        # bounded in time.
        start = time.time()
        mark = LogFileMark('mailman.runner')
        while old_kid in set(m._kids) and time.time() - start < 10:
            time.sleep(0.1)
            m._sigterm_handler(None, None)

        # Check if the runner got the signal.
        start = time.time()
        needle = "command runner caught SIGTERM.  Stopping."
        while (needle not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)
        self.assertIn(needle, mark.read())

        m.thread.join()
        self.assertEqual(len(list(m._kids)), 0)
        m.cleanup()

    def test_sigint_handler(self):
        """Invokes the SIGINT handler.

        We then check whether the runners are actually stopped.
        """
        m = TestableMaster()
        mark = LogFileMark('mailman.runner')
        m.start('command')
        self.assertEqual(len(list(m._kids)), 1)

        kids = list(m._kids)
        self.assertEqual(len(kids), 1)
        old_kid = kids[0]

        # We need to give the runner some time to install the signal
        # handler.  If we send SIGINT before it does, the default
        # action is to terminate the process and will return a
        # slightly different status code.  We wait until we see that
        # the runner has done so by inspecting the log file.  This is
        # race free, and bounded in time.
        start = time.time()
        while ("runner started." not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)

        # Invoke the handler in a loop.  This is race free, and
        # bounded in time.
        start = time.time()
        mark = LogFileMark('mailman.runner')
        while old_kid in set(m._kids) and time.time() - start < 10:
            time.sleep(0.1)
            m._sigint_handler(None, None)

        # Check if the runner got the signal.
        start = time.time()
        needle = "command runner caught SIGINT.  Stopping."
        while (needle not in mark.read()
               and time.time() - start < 10):
            time.sleep(0.1)
        self.assertIn(needle, mark.read())

        m.thread.join()
        self.assertEqual(len(list(m._kids)), 0)
        m.cleanup()

    def test_runner_restart_on_sigkill(self):
        """Kills a runner with SIGKILL and see if it restarted."""
        m = TestableMaster()
        m._restartable = True
        m.start('command')
        kids = list(m._kids)
        self.assertEqual(len(kids), 1)
        old_kid = kids[0]

        # Kill it.  No need to wait for anything as this cannot be
        # caught anyway.
        os.kill(old_kid, signal.SIGKILL)

        # But, we need to wait for the master to collect it.
        start = time.time()
        while old_kid in set(m._kids) and time.time() - start < 10:
            time.sleep(0.1)

        new_kids = list(m._kids)
        self.assertEqual(len(new_kids), 1)
        self.assertTrue(old_kid != new_kids[0])
        m._restartable = False
        for pid in new_kids:
            try:
                os.kill(pid, signal.SIGTERM)
            except ProcessLookupError:
                pass
        m.thread.join()
        m.cleanup()
